import bpy
import bmesh
from copy import copy
from mathutils import Vector, Matrix, Quaternion
from math import degrees, acos

class TestPanel(bpy.types.Panel):
    bl_idname = "object.match_clone_rot"
    bl_label = "MM Tools"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'

    def draw(self, context):
        layout = self.layout
        layout.operator("object.match_clone_rot", text="Match Rotation")

def matchMeshStats(sourceObject):
    #Returns aall meshes that match the vert, face, and edge of <sourceObject>
    mesh = sourceObject.data
    faceCount = len(mesh.polygons)
    edgeCount = len(mesh.edges)
    vertCount = len(mesh.vertices)
    
    #List of matching meshes
    matchedMeshes = []
    
    for element in bpy.context.scene.objects:
        
        # Checks to see if the mesh is hidden (will skip those)
        if element.hide:
            continue
        if element.type != "MESH":
            continue #Skip non meshes
        if element == sourceObject:
            continue #Skip Self
        if len(element.data.vertices) == vertCount:
            if len(element.data.edges) == edgeCount:
                if len(element.data.polygons) == faceCount:
                    matchedMeshes.append(element)
        
    return matchedMeshes

def faceNormalToWorldSpace(obj, faceIndex=0):
    # Returns the normal on obj.face[<faceIndex>] in world space
    assert(faceIndex > -1)
    bm = bmesh.new()
    scene = bpy.context.scene
    bm.from_mesh(obj.data)
    bm.normal_update()
    bm.faces.ensure_lookup_table()
    normal = bm.faces[faceIndex].normal
    wmtx = obj.matrix_world
    world_normal = wmtx.to_3x3().inverted().transposed() * normal
    bm.free()
    world_normal.normalize()
    return world_normal

def deselectAllFacesBut(polyObject, faceIndex = 0):
    #Deselects everything on the target object, except <faceIndex>
    #Returns the face selected
    if bpy.context.object.mode == 'EDIT': #put in object mode if needed
        bpy.ops.object.mode_set(mode='OBJECT')
    #Select polyObject
    bpy.context.scene.objects.active = polyObject
    # deselect everything
    bpy.ops.object.mode_set(mode = 'EDIT')
    bpy.ops.mesh.select_all(action = 'DESELECT')
    # select wanted face
    face = polyObject.data.polygons[faceIndex]
    face.select = True
    #back to object mode
    bpy.ops.object.mode_set(mode = 'OBJECT')
    return face

def rotate_object_around_point(obj, angle, direction, point):
    # rotate_object(obj, angle, direction, point)
    #
    # obj - a Blender object
    # angle - angle in radians
    # direction - axis to rotate around (a vector from the origin)
    # point - point to rotate around (a vector)
    R = Matrix.Rotation(angle, 4, direction)
    T = Matrix.Translation(point)
    M = T * R * T.inverted()
    obj.location = M * obj.location
    obj.rotation_euler.rotate(M)
    

class AddCloneObject(bpy.types.Operator):
    bl_idname = "object.match_clone_rot"
    bl_label = "Match Rotation"
    
    def execute(self, context):
        
        #Verify that we have an object selected
        if bpy.context.active_object:
            
            #Verify that the selection is a mesh
            if bpy.context.active_object.type != "MESH":
                self.report({'INFO'}, "Can only run on mesh objects")
                return {'FINISHED'}
            
            #Save the active_object to select after
            
            #Get the matched meshes
            #Also: save the active_object to select after
            originMesh = bpy.context.active_object 
            matchedMeshes = matchMeshStats(originMesh)
            print("Matches (%d): %s" % (len(matchedMeshes), str(matchedMeshes)))
            
            #Deselect all other faces
            originFace = deselectAllFacesBut(originMesh)
            
            #Select our comparison face
            compFace = 0; #Index of the polygon to match
            originFace = originMesh.data.polygons[compFace]
            originFace.select = True
            
            #Get the baseline normal vector
            originPos = originMesh.location
            
            originFaceNorm = copy(faceNormalToWorldSpace(originMesh, compFace))
            #originFaceNorm = copy(originFace.normal) #Local Normal
            
            #print(originMesh.name + " origin normal " + str(originFaceNorm))
            
            #Get reference face local position
            bm = bmesh.new()
            bm.from_mesh(originMesh.data)
            selected_faces = [f for f in bm.faces if f.select]
            for face in selected_faces:
                face_location = face.calc_center_median()
                loc_world_space = originMesh.matrix_world * Vector(face_location)
                rotCenter = copy(loc_world_space)
                
                #The origFaceOffset allows us to stack all of the matched faces at one point
                origFaceOffset = originPos - loc_world_space
                break
            bm.to_mesh(originMesh.data)
            bm.free()
            
            #Process each mesh
            for m in matchedMeshes:
                mOrigPos = copy(m.location)
                #m.location = originPos
                
                #Make the rotated object's name
                print(m.name)
                
                #Deselect all other faces
                deselectAllFacesBut(m)
                
                #Select our comparison face
                mFace = m.data.polygons[compFace]
                mFace.select = True
                
                #Get face local position (so all of these matched polys are one the same point)
                bm = bmesh.new()
                bm.from_mesh(m.data)
                selected_faces = [f for f in bm.faces if f.select]
                for face in selected_faces:
                    face_location = face.calc_center_median()
                    loc_world_space = m.matrix_world * Vector(face_location)
                    
                    offset = mOrigPos - loc_world_space
                    #print ("Dest offset:" + str(offset))
                    break
                bm.to_mesh(m.data)
                bm.free()
                
                #Position the object so that its comapired 
                m.location = originPos + (offset-origFaceOffset)
                
                #Now get the relative offset of the normal vectors
                mNormal = copy(faceNormalToWorldSpace(m, compFace))
                print ("Start Normal: " + str(mNormal))
                #mNormal = copy(mFace.normal)
                
                #Dont rotate if the normal vectors match check
                if (mNormal != originFaceNorm):
                
                    #Calculate rotation based off of one axis
                    z = Vector(mNormal)
                    
                    #Used Method
                    method = 1
                    
                    if method == 1:
                        #Get Angle For Rotation
                        theRotAxis = z.cross(Vector(originFaceNorm))
                        theRotAxis.normalize()
                        direction = theRotAxis
                        angle = acos(z.dot(originFaceNorm))
                        
                        #Set Cursor and Rotate
                        bpy.context.scene.cursor_location = rotCenter
                        rotate_object_around_point(m, angle, direction, rotCenter)
                    
                    #Update Context (needed for the following print)
                    bpy.context.scene.update()
                    
                    #Print Results to Compaire
                    print ("Final Normal: " + str(faceNormalToWorldSpace(m, compFace)))
                    print ("Reference Normal: "  + str(originFaceNorm))
                    
            
            #Move cursor to pivot point
            print("PP: " + str(rotCenter))
            bpy.context.scene.cursor_location = rotCenter
            bpy.context.scene.objects.active = originMesh
            
            self.report({'INFO'}, "Monkeys are moved")
            
            return {'FINISHED'}
        
        else:
            self.report({'INFO'}, "No object selected to match")
            return {'FINISHED'}

def register():
    bpy.utils.register_module(__name__)
    print("-----"*15)

def unregister():
    bpy.utils.unregister_module(__name__)

if __name__ == "__main__":
    register()
    
    """
    #elif method == 2:
                    #    pass
                        # Calculate the rotation difference between the selected and active faces' normals
                        #mat_rot = z.rotation_difference( originFaceNorm ).to_matrix()
                        #mat_trans = Matrix.Translation(rotCenter)
                        #rotate_with_matrix(m, theTM) #Empty function
                        
                        #Apply rotation
                        #m.rotation_euler[0] = rot[0]
                        #m.rotation_euler[1] = rot[1]
                        #m.rotation_euler[2] = rot[2]
                        #m.location = mOrigPos
                        """